# Extracts a singla RAW depth file of contiguous data
#
# Requires to build xndec
import struct
from onifile import *
from xndec import *

if __name__ == "__main__":
    import sys,os

    if len(sys.argv) == 0:
        print "Required: ONIfilename"
        sys.exit(-1)

    docolor = len(sys.argv) > 3

    filesize = os.stat(sys.argv[1]).st_size
    a = open(sys.argv[1],"rb")
    h0 = readhead1(a)


    mid = 0 # maximum identifier
    prelast = None 
    last = None
    st = None

    offdict = dict()
    count = 0
    ob = allocoutput16(512*424)
    # scan all and keep pre and last
    while True:
            h = readrechead(a)
            if h is None:
                    break
            prelast = last
            last = h
            if h["nid"] > mid:
                    mid = h["nid"]
            if h["nid"] == 1:
                if h["rt"] == RECORD_NEW_DATA:
                    pd = parsedata(a,h)
                    print pd["dataoffset"],h["ps"],h["fs"]
                    z = a.read(h["ps"])
                    count += 1
                    if count == 50:
                        code,size = doXnStreamUncompressDepth16ZWithEmbTable(z,ob)
                        print "decoded ",code,size,"vs input",len(z),"output",len(ob)

                        o = open("x.depth","wb")
                        o.write(ob)
                        o.close()
                        break
            if h["rt"] == RECORD_END:
                continue
            a.seek(h["nextheader"],0) #h["fs"]-HEADER_SIZE+h["ps"],1)

    a.close()



from . import onifile as oni
import struct
import binascii

def info(args,a):
    r = oni.Reader(a)
    print "fileheader",r.h0
    while True:
        h = r.next()
        if h is None:
            break
        elif h["rt"] == oni.RECORD_NEW_DATA:
            break
    for k,v in r.streams.iteritems():
        print k,v

def seeks(args,a):
    """prints seeks"""
    r = oni.Reader(a)
    while True:
        h = r.next()
        if h is None:
            break
        if h["rt"] == oni.RECORD_SEEK_TABLE:
            q = oni.parseseek(a,h)
            for qq in q["data"]:
                print h["nid"],qq

def times(args,a):
    """prints times info"""
    r = oni.Reader(a)
    while True:
        h = r.next()
        if h is None:
            break
        if h["rt"] == oni.RECORD_NEW_DATA:
            hh = oni.parsedatahead(a,h)
            print h["nid"],hh["frameid"],hh["timestamp"]

def compare(args,action,a,b):
    ra = oni.Reader(a)
    rb = oni.Reader(b)
    while True:
        ha = ra.next()
        hb = rb.next()
        if ha != hb:
            print "mismatch in header a=",ha,"b=",hb
        elif ha is None:
            print "ended",ha,hb
            break
        # content
        a.seek(ha["poffset"])
        da = a.read(ha["fs"]-oni.HEADER_SIZE)
        b.seek(hb["poffset"])
        db = b.read(hb["fs"]-oni.HEADER_SIZE)
        if da != db:
            print "mismatch in header extra a=",ha,"b=",hb,len(da),len(db)
        else:
            a.seek(ha["poffset"]+ha["fs"])
            da = a.read(ha["ps"])
            b.seek(ha["poffset"]+ha["fs"])
            db = b.read(ha["ps"])
            if da != db:
                print "mismatch in payload a=",ha,"b=",hb,len(da),len(db)
                if ha["rt"] == oni.RECORD_SEEK_TABLE:
                    oni.parseseek(a,ha)
                    oni.parseseek(b,hb)
                    print "firstseek","\n\t".join([str(x) for x in ha["data"][0:10]])
                    print "secondseek","\n\t".join([str(x) for x in hb["data"][0:10]])
                    #print seek tables
            #print binascii.hexlify(da)
            #print binascii.hexlify(db)

def checkregistered(args,a):
    r = oni.Reader(a)
    last = None
    # scan all and keep pre and last
    mid = -1
    found = False
    while True:
        h = r.next()
        pt = a.tell()
        if h is None:
            break
        prelast = last
        last = h
        if mid < h["nid"]:
            mid = h["nid"]
        elif h["rt"] == oni.RECORD_NODE_DATA_BEGIN:
            break
        elif h["rt"] == oni.RECORD_INT_PROPERTY:
            q = oni.parseprop(a,h)
            if q["name"] == "RegistrationType":
                print q["data"]
                #print q,h,h["poffset"]+2+len("RegistrationType")+4 # namelen[2] name[namelen] datalen[4] data[]
                found = True
                break
    if not found:
        print "RegistrationType MISSING"

def dump(args,a):
    r = oni.Reader(a)
    last = None
    # scan all and keep pre and last
    mid = -1
    print "fileheader",r.h0
    while True:
        h = r.next()
        pt = a.tell()
        if h is None:
            break
        prelast = last
        last = h
        if mid < h["nid"]:
            mid = h["nid"]
        if h["rt"] == oni.RECORD_NEW_DATA:
            print "newdata",h["nid"],oni.parsedatahead(a,h),h["ps"]
        elif h["rt"] == oni.RECORD_NODE_DATA_BEGIN:
            print "nodebegin",h["nid"],h
        elif h["rt"] == oni.RECORD_NODE_ADDED:
            print "nodeadded",h["nid"],h
            print "\tnodeadded-details",oni.parseadded(a,h)
        elif h["rt"] == oni.RECORD_INT_PROPERTY:
            print "intprop",h["nid"],oni.parseprop(a,h),h["poffset"],h["fs"]
        elif h["rt"] == oni.RECORD_REAL_PROPERTY:
            print "realprop",h["nid"],oni.parseprop(a,h),h["poffset"],h["fs"]
        elif h["rt"] == oni.RECORD_GENERAL_PROPERTY:
            pp = oni.parseprop(a,h)
            print "genprop",h["nid"],pp
            if pp["name"] == "xnMapOutputMode":
                (xres,yres) = struct.unpack("ii",pp["data"])
                data = dict(xres=xres,yres=yres)
                print "\t",data
        elif h["rt"] == oni.RECORD_NODE_STATE_READY:
            print "ready",h
        elif h["rt"] == oni.RECORD_NODE_REMOVED:
            print "removed",h
        elif h["rt"] == oni.RECORD_SEEK_TABLE:
            print "seektable",h
        elif h["rt"] == oni.RECORD_END:
            print "end",h
        else:
            print h["rt"]
    if True:
        print "blocks stats:"
        print "\tprelast",prelast
        print "\tlast",last
        print "\tmaxnodeid",mid   
    print "streams stats:"
    for k,v in r.streams.iteritems():
        print "\t",k,v



#!/usr/bin/python
# OpenNI ONI tool
# by Emanuele Ruffaldi PERCRO-SSSA 2013-2015
#
# Implemented Operations:
# - time rescale
# - print times
# - fix truncated file
# - extract motion jpeg of color
# - remove color component
# - extract single image/depth
#
# Possible future:
# - IN PROGRESS split/cut
# - IN PROGRESS 
# - dump stats
#
# TODO: XN_STREAM_PROPERTY_S2D_TABLE
#
# Version 2015/06/18

import struct#
import onitool.onifile as oni
import shutil
import array
import sys
import onitool.toolext as toolext
try:
    import onitool.toolreg as toolreg
except:
    print sys.exc_info()
    toolreg =None
import onitool.toolinfo as toolinfo
import onitool.toolcut as toolcut
import onitool.tooltime as tooltime
import onitool.toolfix as toolfix
try:
    import onitool.toolreg as toolreg
except:
    toolreg = None
    print "missing dll for toolreg"


def interval(s,t,tt):
    try:
        x, y = map(tt, s.split(','))
        return x, y
    except:
        raise argparse.ArgumentTypeError("%s must be x,y" % t)

if __name__ == "__main__":
    import sys,os,argparse

    parser = argparse.ArgumentParser(description='Scan OpenNI ONI files')
    parser.add_argument('file',help='ONI filename')
    parser.add_argument('--info',action="store_true")
    parser.add_argument('--times',action="store_true",help="print times of each frame")
    parser.add_argument('--seeks',action="store_true",help="print seeks")
    parser.add_argument('--dump',action="store_true")
    parser.add_argument('--copy',action="store_true",help="simply copies the content of the ONI via parsing (rebuilds the seektable only)")
    parser.add_argument('--rescale',type=float,default=0.0,help="rescale timings")
    parser.add_argument('--fixcut',action="store_true",help="fixes cut file NOT TESTED HERE")
    parser.add_argument('--fixnite',action="store_true",help="fixes nite file NOT TESTED HERE")
    parser.add_argument('--checkcut',action="store_true",help="checks if file was cut NOT TESTED HERE")
    parser.add_argument('--dupframes',type=int,default=None,help="duplicate frames")
    parser.add_argument('--stripcolor',action="store_true")
    parser.add_argument('--stripdepth',action="store_true")
    parser.add_argument('--noseek',action="store_true",help="disable seek")
    parser.add_argument('--compare',help="compare another file")
    parser.add_argument('--extractframes',type=str,help="extract frames from the given file, requires,")
    parser.add_argument('--stripir',action="store_true",help="removes IR")
    parser.add_argument('--no-registersynctime',dest="registersynctime",action="store_false",help="no sync times while registering")
    parser.add_argument('--registercolor',action="store_true",help="registers color over depth (output is depth size)")
    parser.add_argument('--registerdepth',action="store_true",help="registers depth over color (output is color size, uses filter)")
    parser.add_argument('--mjpeg',action="store_true",help="extract the color stream as motion jpeg")
    parser.add_argument('--extractcolor',help="extract the color stream single jpeg or png images. This option specifies the target path, numbering is the frame number")
    parser.add_argument('--extractdepth',help="extract the depth stream single png images. This option specifies the target path, numbering is the frame number")
    parser.add_argument('--extractir',help="extract the depth stream single png images. This option specifies the target path, numbering is the frame number")
    parser.add_argument('--fseek',type=int,default=0,help="seek frame for extract")
    parser.add_argument('--fduration',type=int,default=-1,help="duration of extraction in frames")
    parser.add_argument('--skipframes',type=int,default=None,help="skip n frames")
    parser.add_argument('--coloreddepth',action="store_true",help="colored depth")
    parser.add_argument('--makeregistered',action="store_true",help="fake reg")
    parser.add_argument('--registered',type=int,default=-1,help="registered modified for fixnite")
    parser.add_argument('--isregistered',action="store_true",help="check registered")
    parser.add_argument('--output',default=None)
    parser.add_argument('--outputdir')
    #parser.add_argument('--extractboth',help="extract the depth and color stream as png and jpeg respectively. This option specifies the target path, numbering is the frame number")
    parser.add_argument('--cutbytime',help="cut by specifing time in seconds: startseconds,endseconds",type=lambda x:interval(x,"time",float))
    parser.add_argument('--cutbyframe',help="cut by specifing time in frames: startframe,endframe",type=lambda x:interval(x,"frame",int))

    args = parser.parse_args()

    if args.file is None:
        print "Required: ONI filename"
        sys.exit(-1)

    action = ""
    patchaction = False



    if args.mjpeg:
        if action != "":
            print "Already specified action",action
            sys.exit(-1)
        if args.output is None:
            print "Required: ONI filename in output --output"
            sys.exit(-1)
        action = "mjpeg"

    if args.rescale:
        if action != "":
            print "Already specified action",action
            sys.exit(-1)
        if args.output is None:
            print "Required: ONI filename in output --output"
            sys.exit(-1)
        action = "rescale"

    if args.stripcolor:
        if action != "":
            print "Already specified action",action
            sys.exit(-1)
        if args.output is None:
            print "Required: ONI filename in output --output"
            sys.exit(-1)
        action = "stripcolor"

    if args.stripdepth:
        if action != "":
            print "Already specified action",action
            sys.exit(-1)
        if args.output is None:
            print "Required: ONI filename in output --output"
            sys.exit(-1)
        action = "stripdepth"

    if args.stripir:
        if action != "":
            print "Already specified action",action
            sys.exit(-1)
        if args.output is None:
            print "Required: ONI filename in output --output"
            sys.exit(-1)
        action = "stripir"

    if args.fixcut :
        if action != "":
            print "Already specified action",action
            sys.exit(-1)
        if args.output is not None:
            print "FIXCUT works in patch mode"
            sys.exit(-1)
        patchaction = True
        action = "fixcut"

    if args.makeregistered:
        if action != "":
            print "Already specified action",action
            sys.exit(-1)
        if args.output is None:
            print "FIXNITE requires output"
            sys.exit(-1)
        action = "makeregistered"

    if args.isregistered:
        if action != "":
            print "Already specified action",action
            sys.exit(-1)
        action = "isregistered"
    if args.fixnite :
        if action != "":
            print "Already specified action",action
            sys.exit(-1)
        if args.output is None:
            print "FIXNITE requires output"
            sys.exit(-1)
        action = "fixnite"

    if args.copy:
        if action != "":
            print "Already specified action",action
            sys.exit(-1)
        if args.output is None:
            print "Required: ONI filename in output --output"
            sys.exit(-1)
        action = "copy"

    if args.seeks:
        if action != "":
            print "Already specified action",action
            sys.exit(-1)
        if args.output is not None:
            print "Not Required output"
            sys.exit(-1)
        action = "seeks"

    if args.checkcut:
        if action != "":
            print "Already specified action",action
            sys.exit(-1)
        if args.output is not None:
            print "Not Required: ONI filename in output --output"
            sys.exit(-1)
        action = "checkcut"

    if args.cutbyframe is not None:
        if action != "":
            print "Already specified action",action
            sys.exit(-1)
        if args.output is None:
            print "Required: ONI filename in output --output"
            sys.exit(-1)
        target = ("frame",args.cutbyframe)
        action = "cutbyframe"

    if args.cutbytime is not None:
        if action != "":
            print "Already specified action",action
            sys.exit(-1)
        if args.output is None:
            print "Required: ONI filename in output --output"
            sys.exit(-1)
        target = ("timeus",[int(x*1E6) for x in args.cutbyframe])
        action = "cutbytime"

    if args.extractframes:
        if action != "":
            print "Already specified action",action
            sys.exit(-1)
        if args.outputdir == "":
            print "output dir needed for extractframes"
            sys.exit(-1)
        action = "extractframes"



    subaction = ""
    extractpath = ""
    varia = ['extractcolor','extractdepth','extractir'] #,'extractboth']
    for x in varia:
        if getattr(args,x):
            if action != "":
                print "Already specified action",action
                sys.exit(-1)
            subaction = x
            action = "extract"
            extractpath = getattr(args,x)
            print "extract action",extractpath

    if args.dump:
        if action != "":
            print "Already specified action",action
            sys.exit(-1)
        action = "dump"

    if args.info:
        if action != "":
            print "Already specified action",action
            sys.exit(-1)
        action = "info"

    if args.compare is not None:
        if action != "":
            print "Already specified action",action
            sys.exit(-1)
        action = "compare"

    if args.dupframes:
        if action != "":
            print "Already specified action",action
            sys.exit(-1)
        if args.output is None:
            print "Required: ONI filename in output --output"
            sys.exit(-1)
        action = "dupframes"

    if args.skipframes:
        if action != "":
            print "Already specified action",action
            sys.exit(-1)
        action = "skipframes"
    if args.registercolor:
        if action != "":
            print "Already specified action",action
            sys.exit(-1)
        if args.output is None:
            print "Required: ONI filename in output --output"
            sys.exit(-1)
        action = "registercolor"
    if args.registerdepth:
        if action != "":
            print "Already specified action",action
            sys.exit(-1)
        if args.output is None:
            print "Required: ONI filename in output --output"
            sys.exit(-1)
        action = "registerdepth"

    filesize = os.stat(args.file).st_size
    if patchaction:
        shutil.copyfile(args.file,args.output)
        args.file = args.output
        print args.file
        a = open(args.file,"r+b")
    else:
        a = open(args.file,"rb")
        if action != "" and args.output is not None:
            b = open(args.output,"wb")

    if action == "registercolor" or action == "registerdepth":
        toolreg.register(args,action,a,b)
    elif action == "extract":
        toolext.extract(args,subaction,extractpath,a)        
    elif action == "isregistered":
        toolinfo.checkregistered(args,a)
    elif action == "mjpeg":
        toolext.extractmjpeg(args,a,b)        
    elif action == "stripcolor" or action == "stripir" or action == "stripdepth":
        toolcut.strip(args,action,a,b)
    elif action == "cutbyframe" or action == "cutbytime":
        toolcut.cut(args,target,a,b)
    elif action == "rescale":
        tooltime.rescale(args,args.rescale,a,b)
    elif action == "copy":
        toolcut.copy(args,a,b)
    elif action == "dupframes":
        toolcut.dupframes(args,a,b)
    elif action == "seeks":
        toolinfo.seeks(args,a)
    elif action == "skipframes":
        toolcut.skip(args,a,b)
    elif action == "fixcut" or action == "checkcut":
        toolfix.fix(args,action,a)
    elif action == "fixnite":
        toolfix.fixnite(args,action,a,b)
    elif action == "compare":
        toolinfo.compare(args,action,a,open(args.compare,"rb"))
    elif action == "makeregistered":
        toolfix.makeregistered(args,action,a,b)
        #args.registered = 2
        #toolfix.fixnite(args,action,a,b)
    elif args.times:
        toolinfo.times(args,a)
    elif action == "info":
        toolinfo.info(args,a)
    elif action == "extractframes":
        toolext.extractframes(args,a,[int(x) for x in open(args.extractframes,"r")],args.outputdir)
    elif action == "dump" or action == "":
        toolinfo.dump(args,a)
    else:
        print "unknown action",action

    if False:
        if needclose:
            # if need to close first PATCH the NODE ADDED
            # then patch the writeseek        
            for q in stats.values():
                print "writing",q
                q.patchframeheader(b)
                q.writeseek(b)
            h0["ts"] = max([q.maxts for q in stats.values()])
            oni.writehead1(b,h0)                           
        a.close()




from . import onifile as oni
import struct
from collections import defaultdict


def rescale(args,factor,a,b):
    r = oni.Reader(a)
    w = oni.Writer(b,r.h0)
    while True:
        h = r.next()
        if h is None:
            break
        elif h["rt"] == oni.RECORD_SEEK_TABLE:
            w.emitseek(h["nid"])
        elif h["rt"] == oni.RECORD_NEW_DATA:
            hh = oni.parsedatahead(a,h)            
            print dict(nid=h["nid"],ps=h["ps"],fs=h["fs"],frameid=hh["frameid"],timestamp=hh["timestamp"])
            w.addframe(h["nid"],hh["frameid"],factor*hh["timestamp"],a.read(h["ps"]))
        else:
            w.copyblock(h,a)
    w.finalize()

"""
        # TODO update with StreamInfo
        stats = dict()
        while True:
                h = oni.readrechead(a)
                if h is None:
                        break
                if h["rt"] == oni.RECORD_NEW_DATA:
                    t = oni.parsedatahead(a,h)["timestamp"]
                    t2 = int(t*args.rescale)
                    #print h["nid"],t,t2
                    q = stats.get(h["nid"],None)
                    if q is None:
                        print "new for",h["nid"],"in new data"
                        q = [t2,t2,None,None]
                        stats[h["nid"]] = q
                    else:
                        if q[0] is None:
                            q[0] = t2
                            q[1] = t2
                        else:
                            if q[0] < t2:
                                q[0] = t2;
                            if q[1] > t2:
                                q[1] = t2
                    oni.patchtime(a,h,t2)
                elif h["rt"] == oni.RECORD_INT_PROPERTY:
                    z = oni.parseprop(a,h)
                    if z["name"] == "IsFrameBased":
                        z["data"] = 0
                        oni.writeprop(a,h,z)
                elif h["rt"] == oni.RECORD_NODE_ADDED:
                    hh = oni.parseadded(a,h)
                    q = stats.get(h["nid"],None)
                    if q is None:
                        print "new for",h["nid"],"in added"
                        q = [None,None,h,hh]
                        stats[h["nid"]] = q
                    else:
                        q[2] = h
                        q[3] = hh
                elif h["rt"] == oni.RECORD_SEEK_TABLE:
                    x = oni.parseseek(a,h)
                    n = len(x["data"])
                    print "seektable",h["nid"],"of",n
                    for i in range(0,n):
                        x["data"][i]["timestamp"] = int(x["data"][i]["timestamp"]*args.rescale)
                    oni.writeseek(a,h) 
                    
                # next record
                a.seek(h["nextheader"]) #h["fs"]-HEADER_SIZE+h["ps"]+pt,0)      
        needclose = True
"""




import ctypes,os

# XnStatus XnStreamUncompressDepth16ZWithEmbTable(const XnUInt8* pInput, const XnUInt32 nInputSize, XnUInt16* pOutput, XnUInt32* pnOutputSize)
# XnStatus XnStreamUncompressDepth16Z(const XnUInt8* pInput, const XnUInt32 nInputSize, XnUInt16* pOutput, XnUInt32* pnOutputSize)
# XnStatus XnStreamUncompressImage8Z(const XnUInt8* pInput, const XnUInt32 nInputSize, XnUInt8* pOutput, XnUInt32* pnOutputSize)
# XnStatus XnStreamUncompressConf4(const XnUInt8* pInput, const XnUInt32 nInputSize, XnUInt8* pOutput, XnUInt32* pnOutputSize)

types = ["",".so",".dylib",".pyd"]
x = None
path = os.path.abspath(__file__)
dir_path = os.path.dirname(path)

for a in types:
	try:
		x = ctypes.CDLL(os.path.join(dir_path,"libxndec" + a))
		break
	except:
		pass
f = x.XnStreamUncompressDepth16ZWithEmbTable
et = ctypes.POINTER(ctypes.c_uint8)
f.argtypes = [ctypes.c_char_p,ctypes.c_int,ctypes.POINTER(ctypes.c_uint16),ctypes.POINTER(ctypes.c_int)]
f.restype = ctypes.c_int
XnStreamUncompressDepth16ZWithEmbTable = f

f = x.XnStreamUncompressDepth16Z
et = ctypes.POINTER(ctypes.c_uint8)
f.argtypes = [ctypes.c_char_p,ctypes.c_int,ctypes.POINTER(ctypes.c_uint16),ctypes.POINTER(ctypes.c_int)]
f.restype = ctypes.c_int
XnStreamUncompressDepth16Z = f

f = x.XnStreamUncompressImage8Z
et = ctypes.POINTER(ctypes.c_uint8)
f.argtypes = [ctypes.c_char_p,ctypes.c_int,ctypes.POINTER(ctypes.c_uint8),ctypes.POINTER(ctypes.c_int)]
f.restype = ctypes.c_int
XnStreamUncompressImage8Z = f

f = x.XnStreamUncompressConf4
et = ctypes.POINTER(ctypes.c_uint8)
f.argtypes = [ctypes.c_char_p,ctypes.c_int,ctypes.POINTER(ctypes.c_uint8),ctypes.POINTER(ctypes.c_int)]
f.restype = ctypes.c_int
XnStreamUncompressConf4 = f

#XnStatus XnStreamCompressDepth16ZWithEmbTable(const XnUInt16* pInput, const XnUInt32 nInputSize, XnUInt8* pOutput, XnUInt32* pnOutputSize, XnUInt16 nMaxValue)
f = x.XnStreamCompressDepth16ZWithEmbTable
et = ctypes.POINTER(ctypes.c_uint8)
f.argtypes = [ctypes.POINTER(ctypes.c_uint16),ctypes.c_int,ctypes.POINTER(ctypes.c_uint8),ctypes.POINTER(ctypes.c_int),ctypes.c_int]
f.restype = ctypes.c_int
XnStreamCompressDepth16ZWithEmbTable = f


def allocoutput16(n):
	#pt = ctypes.c_uint16*n
	#p = pt()
	#return p
	return ctypes.create_string_buffer(n*2)

def allocoutput8(n):
	#pt = ctypes.c_uint16*n
	#p = pt()
	#return p
	return ctypes.create_string_buffer(n)

def doXnStreamUncompressConf4(input,outputbuffer):
	r = ctypes.c_int(len(outputbuffer));
	rr = XnStreamUncompressConf4(ctypes.c_char_p(input),len(input),ctypes.cast(outputbuffer,ctypes.POINTER(ctypes.c_uint8)),ctypes.byref(r))
	return (rr,r.value)

def doXnStreamUncompressImage8Z(input,outputbuffer):
	r = ctypes.c_int(len(outputbuffer));
	rr = XnStreamUncompressImage8Z(ctypes.c_char_p(input),len(input),ctypes.cast(outputbuffer,ctypes.POINTER(ctypes.c_uint8)),ctypes.byref(r))
	return (rr,r.value)

def doXnStreamUncompressDepth16ZWithEmbTable(input,outputbuffer):
	r = ctypes.c_int(len(outputbuffer));
	rr = XnStreamUncompressDepth16ZWithEmbTable(ctypes.c_char_p(input),len(input),ctypes.cast(outputbuffer,ctypes.POINTER(ctypes.c_uint16)),ctypes.byref(r))
	return (rr,r.value)

def doXnStreamCompressDepth16ZWithEmbTable(input,outputbuffer,maxvalue):
	r = ctypes.c_int(len(outputbuffer));
	rr = XnStreamCompressDepth16ZWithEmbTable(ctypes.cast(input,ctypes.POINTER(ctypes.c_uint16)),len(input),ctypes.cast(outputbuffer,ctypes.POINTER(ctypes.c_uint8)),ctypes.byref(r),maxvalue)
	return (rr,r.value)


def doXnStreamUncompressDepth16Z(input,outputbuffer):
	r = ctypes.c_int(len(outputbuffer));
	rr = XnStreamUncompressDepth16Z(ctypes.c_char_p(input),len(input),ctypes.cast(outputbuffer,ctypes.POINTER(ctypes.c_uint16)),ctypes.byref(r))
	return (rr,r.value)


if __name__ == "__main__":

	a = "1234"
	b = allocoutput16(123)
	print doXnStreamUncompressDepth16ZWithEmbTable(a,b)


import struct,sys
from collections import defaultdict
RECORD_END = 0x0B
HEADER_MAGIC_SIZE = 4
MAGIC = 0x0052494E
MAGICS = "NIR\0"
RHMAGIC = "NI10"
HEADER_SIZE = 4*5+8

def XN_CODEC_ID(c1, c2, c3, c4):
    c1 = ord(c1)
    c2 = ord(c2)
    c3 = ord(c3)
    c4 = ord(c4)
    return ((c4 << 24) | (c3 << 16) | (c2 << 8) | c1)


NODE_TYPE_DEVICE = 1
NODE_TYPE_DEPTH = 2
NODE_TYPE_IMAGE= 3
NODE_TYPE_IR = 4
#NODE_TYPE_CLOUD = 5 # ADDED by ER, means 3 floats of the cloud decomposed, associated to the
#   just the points, then the color is take from IMAGE if registered...
#   otherwise needs an additional (secondary) IMAGE...

XN_CODEC_UNCOMPRESSED =XN_CODEC_ID('N','O','N','E')
XN_CODEC_16Z = XN_CODEC_ID('1','6','z','P')
XN_CODEC_16Z_EMB_TABLES =XN_CODEC_ID('1','6','z','T')
XN_CODEC_8Z =XN_CODEC_ID('I','m','8','z')
XN_CODEC_JPEG =XN_CODEC_ID('J','P','E','G')


ONI_PIXEL_FORMAT_DEPTH_1_MM = 100
ONI_PIXEL_FORMAT_DEPTH_100_UM = 101
ONI_PIXEL_FORMAT_SHIFT_9_2 = 102
ONI_PIXEL_FORMAT_SHIFT_9_3 = 103

ONI_PIXEL_FORMAT_RGB888 = 200
ONI_PIXEL_FORMAT_YUV422 = 201
ONI_PIXEL_FORMAT_GRAY8 = 202
ONI_PIXEL_FORMAT_GRAY16 = 203
ONI_PIXEL_FORMAT_JPEG = 204
ONI_PIXEL_FORMAT_YUYV = 205

RECORD_NODE_ADDED_1_0_0_4       = 0x02
RECORD_INT_PROPERTY             = 0x03
RECORD_REAL_PROPERTY            = 0x04
RECORD_STRING_PROPERTY          = 0x05
RECORD_GENERAL_PROPERTY         = 0x06
RECORD_NODE_REMOVED             = 0x07
RECORD_NODE_DATA_BEGIN          = 0x08
RECORD_NODE_STATE_READY         = 0x09
RECORD_NEW_DATA                 = 0x0A
RECORD_END                      = 0x0B
RECORD_NODE_ADDED_1_0_0_5       = 0x0C
RECORD_NODE_ADDED               = 0x0D
RECORD_SEEK_TABLE               = 0x0E

IMAGE_REGISTRATION_OFF = 0
IMAGE_REGISTRATION_DEPTH_TO_COLOR = 1
"""
struct FileHeaderData
{
    XnUInt8 identity[IDENTITY_SIZE];
    struct Version
    {
        XnUInt8 major;
        XnUInt8 minor;
        XnUInt16 maintenance;
        XnUInt32 build;
    } version;
    XnUInt64 maxTimeStamp;
    XnUInt32 maxNodeId;
};
struct RecordHeaderData
{
    XnUInt32 magic;
    XnUInt32 recordType;
    XnUInt32 nodeId;
    XnUInt32 fieldsSize; // comprises the HEADER_SIZE
    XnUInt32 payloadSize;
    XnUInt64 undoRecordPos;
};

Version structure:
	XnUInt8         nMajor
	XnUInt8         nMinor
	XnUInt16        nMaintenance
	XnUInt32        nBuild
"""

#emitCommonHeader(RECORD_NODE_STATE_READY, nodeId, /*undoRecordPos*/ 0);

#emit_RECORD_NODE_DATA_BEGIN
#   framecount 4
#   maxts 8

#emit_RECORD_NEW_DATA
#   timestamp 8
#    Seek table position

#struct RecordHeaderData
#{
#    XnUInt32 magic;
#    XnUInt32 recordType;
#    XnUInt32 nodeId;
#    XnUInt32 fieldsSize;
#    XnUInt32 payloadSize;
#    XnUInt64 undoRecordPos;
#};


def parseint(a):
    return struct.unpack("i",a.read(4))[0]

def parseint64(a):
    return struct.unpack("Q",a.read(8))[0]

def makeint64(a):
    return struct.pack("Q",a)

def parsedatahead(a,h):
    """Parsed the header of the data block containing timestamp and seek table position
    https://github.com/OpenNI/OpenNI2/blob/master/Source/Core/OniDataRecords.cpp"""
    a.seek(h["poffset"],0)
    ts =    parseint64(a)
    frameid  = parseint(a)
    return dict(timestamp=ts,frameid=frameid)

def writedatahead(a,h,hh):
    a.seek(h["poffset"],0)
    a.write(struct.pack("=qi",hh["timestamp"],hh["frameid"]))

def patchtime(a,h,ot):
    a.seek(h["poffset"],0)
    a.write(makeint64(ot))

def parsestr(a):
    namelen = parseint(a)
    name = a.read(namelen)[0:-1]
    return name


def makeindexentry(a):
    """encodes the DataIndexEntry made of a timestamp, config and offset"""
    if type(a) == tuple:
        return struct.pack("=QiQ",a[0],a[1],a[2])
    else:
        return struct.pack("=QiQ",a["timestamp"],a["config"],a["offset"])

#def writeseek(a,h):
#emitCommonHeader(RECORD_SEEK_TABLE, nodeId, /*undoRecordPos*/ 0);
#empty entry for frame 0
#DataIndexEntry emptyEntry; all zero
#
#payloadSize = total entries
#    a.seek(h["poffset"]+h["fs"],0)
#    for x in h["data"]:
#        y = makeindexentry(x)
#        a.write(y)

def parseindexentry(a):
    """decodes the DataIndexEntry made of a timestamp, config and offset as dictionary"""
    ts,cid,pos = struct.unpack("=QiQ",a.read(20))
    return dict(timestamp=ts,config=cid,offset=pos)

def parseseek(a,h):
    a.seek(h["poffset"]+h["fs"]-HEADER_SIZE,0)
    #print "seek",h["fs"]
    r = []
    n = h["ps"]/20
    #print "reading seektable",n
    for i in range(0,n):
        t = parseindexentry(a)
        t["frameid"] = i
        r.append(t)
    h["data"] = r
    return h
def makestr(s):
    return struct.pack("=i",len(s)+1)+ s+"\x00" 

codec2id = dict(raw=XN_CODEC_UNCOMPRESSED,jpeg=XN_CODEC_JPEG)
codec2id["16z"] = XN_CODEC_16Z
codec2id["8z"] = XN_CODEC_8Z
codec2id["16zt"] = XN_CODEC_16Z_EMB_TABLES

def writedadded(a,h,hh):
    a.seek(h["poffset"],0)
    a.write(makestr(hh["name"]))
    #print "codecback",codec2id.get(hh["codec"],hh["codec"])
    ocodec = codec2id.get(hh["codec"],hh["codec"])
    #print h,hh
    a.write(struct.pack("=iiiQQQ",hh["nodetype"],ocodec,hh["frames"],hh["mints"],hh["maxts"],hh["seektable"]))
def parseadded(a,h):
    a.seek(h["poffset"],0)
    name = parsestr(a)
    nodetype = parseint(a)
    codec = parseint(a)
    nframes = parseint(a)
    mints = parseint64(a)
    maxts = parseint64(a)
    seektable =parseint64(a) # seektable header data
    ocodec = codec
    if codec == XN_CODEC_UNCOMPRESSED:
        codec = "raw"
    elif codec == XN_CODEC_16Z:
         codec = "16z"
    elif codec == XN_CODEC_16Z_EMB_TABLES:
        codec = "16zt"
    elif codec == XN_CODEC_8Z:
        codec = "8z"
    elif codec == XN_CODEC_JPEG:
        codec = "jpeg"
    #print "codecin",codec,ocodec
    return dict(name=name,nodetype=nodetype,codec=codec,frames=nframes,mints=mints,maxts=maxts,seektable=seektable)

def parseprop(a,h):
    a.seek(h["poffset"],0)
    name = parsestr(a)
    datalen = parseint(a)-4
    data = a.read(datalen)
    if h["rt"] == RECORD_INT_PROPERTY:
        if datalen == 8:
            data = struct.unpack("q",data)[0]
        else:
            data = struct.unpack("i",data)[0]
    elif h["rt"] == RECORD_REAL_PROPERTY:  
        #print "realprop",h,datalen
        if datalen == 8:
            data = struct.unpack("d",data)[0]
        else:
            data = struct.unpack("f",data)[0]
    return dict(name=name,data=data,datalen=datalen)

def addprop(a,nid,name,type,value,datalen=0):
    if type == RECORD_INT_PROPERTY:      
        #content after header is:         
        if datalen == 8:
            c = makestr(name) + struct.pack("=iq",4+8,value)
        else:
            c = makestr(name) + struct.pack("=iii",4+4,value,0) # last 0 is dummy for leaving space?
        writehead(a,dict(rt=type,nid=nid,ps=0,fs=HEADER_SIZE+len(c),undopos=0))
        a.write(c)
    elif type == RECORD_GENERAL_PROPERTY:
        c = makestr(name) + struct.pack("=i",4+len(value)) + value
        writehead(a,dict(rt=type,nid=nid,ps=0,fs=HEADER_SIZE+len(c),undopos=0))
        a.write(c)        
    elif type == RECORD_REAL_PROPERTY:      
        c = makestr(name) + struct.pack("=if",4+4,value)
        writehead(a,dict(rt=type,nid=nid,ps=0,fs=HEADER_SIZE+len(c),undopos=0))
        a.write(c)
    else:
        print "prop type unsupported",type
        sys.exit(-1)


def writeprop(a,h,z):
    a.seek(h["poffset"],0)
    if h["rt"] == RECORD_INT_PROPERTY:      
        c = makestr(z["name"]) + struct.pack("=ii",4+4,z["data"])
        a.write(c)
    else:
        print "prop type unsupported",h,z
        sys.exit(-1)


def emptyhead1():
    return dict(magic="NI10",version=(1,0,1,0),maxnid=0,ts=0)

def writehead1(a,h):
    """writes a new header"""
    a.seek(0)
    version = struct.pack("bbhi",*h["version"])
    ts = struct.pack("Q",h["ts"])
    maxnodeid = struct.pack("i",h["maxnid"])
    a.write(h["magic"]+version+ts+maxnodeid)

def readhead1(a):        
    """read the main file header"""
    magic = a.read(HEADER_MAGIC_SIZE)
    version = struct.unpack("bbhi",a.read(2+2+4))
    ts = struct.unpack("Q",a.read(8))[0]
    maxnodeid = struct.unpack("i",a.read(4))[0]
    if magic != RHMAGIC:
            print "bad magic",magic
            return False
    else:
            #print magic,version,"maxnodeid",maxnodeid,ts
            return dict(version=version,maxnid=maxnodeid,ts=ts,magic=magic)


def writeend(a):
    """writes the end record"""
    w = (MAGIC,0x0B,0,HEADER_SIZE,0)
    a.write(struct.pack("5i",*w)+struct.pack("Q",0))

def copyblock(a,h,b,frame=None,timestamp=None):
    a.seek(h["poffset"],0)
    hout = dict(rt=h["rt"],nid=h["nid"],fs=h["fs"],ps=h["ps"],undopos=h["undopos"],poffset=0,hoffset=b.tell(),nextheader=0)
    writehead(b,hout)
    if h["fs"] > HEADER_SIZE:
        # adjust time
        if h["rt"] == RECORD_NEW_DATA and frame is not None:
            # skip real time 
            oldts =    parseint64(a)
            oldframeid  = parseint(a)
            # field or payload?
            b.write(struct.pack("=qi",timestamp,frame))
        else:
            #print "extra",h
            b.write(a.read(h["fs"]-HEADER_SIZE))

    done = 0
    while done < h["ps"]:
        left = h["ps"]-done
        if left > 64*1024:
            left = 64*1024
        b.write(a.read(left))
        done += left
    hout["poffset"] = b.tell()
    hout["nextheader"] = hout["hoffset"] + hout["fs"] + hout["ps"]
    return hout

def writehead(a,h):
    a.write(struct.pack("5i",MAGIC,h["rt"],h["nid"],h["fs"],h["ps"]) + struct.pack("Q",h["undopos"]))

def readrechead(a):
    """read record: the resulting dictionary contains:
    - rt  = recordType
    - nid = identifier/stream
    - fs  = field size
    - ps  = payload size
    - poffset = offset to the content
    - hoffset = offset to the header
    - nextheader = next header
    - h2 = undo record pos (one uint64)

    The header stored is:
    - magic (32bit) 0x0052494E
    - rt
    - nid
    - fs = sizeof(*m_header)
    - ps 
    - undorecord
    """
    p = a.tell()
    h1 = a.read(HEADER_SIZE)
    if h1 == "":
            return None
    magic,rt,nid,fs,ps,undopos = struct.unpack("=5iQ",h1)
    if magic != MAGIC:
            print "bad magic record",magic
            return None
    r = dict(rt=rt,nid=nid,fs=fs,ps=ps,poffset=a.tell(),hoffset=p,nextheader=p+ps+fs,undopos=undopos,magic=magic)
    return r


class StreamInfo:
    """Class for transcoding"""
    def __init__(self):
        self.oldtimestamp = 0 # last old times
        self.oldframes = 0 # original frame
        self.oldbasetime = None # of old

        self.newtimestamp = 0 # last new times
        self.newframes = 0 # current writing frame
        self.maxts = None # of new frame
        self.mints = None # of new frame
        self.configid = 0
        self.emitted = False
        self.removeemitted = False

        self.headerblock = None # ??
        self.headerdata = None # ?? 
        self.framesoffset = [(0,0,0)] # stored seek table (timestamp,offset)
        self.headerseek = None # seek header
    def assignnodeadded(self,h,hh):
        self.headerblock = h
        self.headerdata = hh
    def addframe(self,preoffset,dataheader,file,configid):
        ts = dataheader["timestamp"]
        self.newtime(ts)
        self.framesoffset.append((ts,configid,preoffset))
        self.newframes += 1
    def newtime(self,t):
        if self.maxts is None:
            self.maxts = t
            self.mints = t
        else:
            if t > self.maxts:
                self.maxts = t
            if t < self.mints:
                self.mints = t
        self.newtimestamp = t
    def writeseek(self,a,noseek):
        self.emitted = True
        q = self 
        off = a.tell()  #!!
        if not noseek:
            q.headerseek = dict(rt=RECORD_SEEK_TABLE,
                ps=len(self.framesoffset)*20, #QiQ
                fs=HEADER_SIZE, # standard
                nid=self.headerblock["nid"],
                undopos=0)
            q.headerdata["maxts"] = q.maxts is not None and  q.maxts or 0
            q.headerdata["mints"] = q.mints is not None and  q.mints or 0 
            q.headerdata["frames"] = q.newframes
            if len(self.framesoffset) > 1: # skip 0
                q.headerdata["seektable"] = off 
                writehead(a,self.headerseek)
                # then the rest and we already now size and everything
                for t in self.framesoffset:
                    a.write(makeindexentry(t))
            else:
                q.headerdata["seektable"] = 0
        else:
            q.headerdata["seektable"] = 0
    def fixnodeadded(self,a):
        writedadded(a,self.headerblock,self.headerdata) 

class Reader:
    def __init__(self,file,h0=None):
        self.file = file
        self.lasth = None
        self.nodeinfo = dict()
        self.nodetype2nid = dict()
        if h0 is None:
            h0 = readhead1(self.file)
        self.h0 = h0
        self.pseektable = dict()
        self.pend = None
    def getseektable(self,nid):
        if nid not in self.pseektable:
            return None
        return parseseek(self.file,self.pseektable[nid])
    @property
    def streams(self):
        return self.nodeinfo

    def next(self):
        if self.lasth:
            self.file.seek(self.lasth["nextheader"])
        h = readrechead(self.file)
        self.lasth = h
        if h is None or h["magic"] == 0:
            return None
        if h["rt"] == RECORD_NEW_DATA:
            pass
        elif h["rt"] == RECORD_NODE_ADDED:
            hh = parseadded(self.file,h)
            self.nodeinfo[h["nid"]] = hh
            self.nodetype2nid[hh["nodetype"]] = h["nid"]
        elif h["rt"] == RECORD_GENERAL_PROPERTY:
            pp = parseprop(self.file,h)
            if pp["name"] == "xnMapOutputMode":
                (xres,yres) = struct.unpack("ii",pp["data"])
                hhnode = self.nodeinfo[h["nid"]]
                hhnode["size"] = (xres,yres)
        elif h["rt"] == RECORD_NEW_DATA:
            hh = parsedatahead(self.file,h)
            q = self.nodeinfo[hh["nid"]]
            q["maxts"] = hh["timestamp"]
        elif h["rt"] == RECORD_SEEK_TABLE:
            self.pseektable[h["nid"]] = h
        elif h["rt"] == RECORD_END:
            self.pend = h
        return h

# inplace
class Patcher(Reader):
    def __init__(self,file,h0):
        Reader.__init__(file,h0)
        self.stats = defaultdict(StreamInfo) # for file stats and seek table into b
    def finalize(self):
        for q in self.stats.values():
            #print "writing",q
            q.patchframeheader(self.file)
            q.writeseek(self.file,False)
        self.h0["ts"] = max([q.maxts for q in self.stats.values()])
        writehead1(self.file,self.h0)                           
        writeend(self.file)                    

class Writer:
    def __init__(self,file,h0=None):
        self.file = file
        self.stats = defaultdict(StreamInfo) # for file stats and seek table into b
        self.mid = -1
        self.noseek = False
        self.configid = 0
        if h0 is None:
            self.h0 = emptyhead1()
        else:
            self.h0 = dict()
            self.h0.update(h0)
        writehead1(self.file,self.h0)
        self.endemitted = False
    # adds a property via header and content
    # increment configid as any new property
    def addproperty(self,header,content):
        writehead(self.file,header) 
        header["poffset"] = self.file.tell()       
        writeprop(self.file,header,content)
        self.configid += 1
    # adds a property via target,name,type and value
    # increment configid as any new property
    def addprop(self,nid,name,type,value,datalen=0):
        addprop(self.file,nid,name,type,value,datalen)
        self.configid += 1
    # copy a block from one to another, if it is a property increment the config consequently
    def copyblock(self,header,file):
        if header["nid"] > self.mid:
            self.mid = header["nid"]
        file.seek(header["poffset"]) # got to data
        d = file.read(header["ps"]+header["fs"]-HEADER_SIZE) # ps + fs !

        preoffset = self.file.tell()
        writehead(self.file,header) # same header 
        po = self.file.tell() # save output location
        self.file.write(d) # content fs+ps

        # ANALYZE
        rt = header["rt"]
        if rt == RECORD_NODE_ADDED:
            hh = dict()
            hh.update(header) 
            hh["poffset"] = po
            # build node added for stats and seektable
            hd = parseadded(file,header) # parse
            self.stats[header["nid"]].assignnodeadded(hh,hd)

            #print "adding RECORD_NODE_ADDED to output",hh,hd
        elif rt == RECORD_INT_PROPERTY or rt == RECORD_REAL_PROPERTY or rt == RECORD_GENERAL_PROPERTY:
            self.configid += 1
        elif rt == RECORD_NODE_REMOVED:
            self.stats[header["nid"]].removeemitted = True
        elif rt == RECORD_END:
            self.endemitted = True
        elif rt == RECORD_NEW_DATA:
            q = self.stats[header["nid"]]
            dataheader = parsedatahead(file,header) # parse
            q.addframe(preoffset,dataheader,self.file,self.configid)
    def addframe(self,nid,frameid,timestamp,content):
        if nid > self.mid:
            self.mid = nid
        # basehead is 5*4=20 + 8 pos = 28
        # extra data (frame and timestamp)=12
        h = dict(rt=RECORD_NEW_DATA,nid=nid,fs=28+12,ps=len(content),undopos=0)
        preoffset = self.file.tell()
        writehead(self.file,h)         
        h["poffset"] = self.file.tell()     # for the seektable 
        dataheader = dict(frameid=frameid,timestamp=timestamp) # fs
        writedatahead(self.file,h,hh) # fs content write
        self.file.write(content) # ps

        # add for seektable, add frame
        q = self.stats[h["nid"]]
        q.addframe(preoffset,dataheader,self.file,self.configid)
    def emitseek(self,nid,ofile=None,hofile=None):
        for k,q in self.stats.iteritems():
            if q.headerblock["nid"] == nid and not q.emitted:
                if ofile is not None:
                    # we are using an existing file so copy the config id 
                    pp = parseseek(ofile,hofile)
                    configid = pp["data"][1]["config"]
                    print "newconfigid",configid
                    self.configid = configid
                    for i,o in enumerate(q.framesoffset):
                        q.framesoffset[i] = (o[0],configid,o[2])
                    q.framesoffset[0] = (0,0,0)
                #print "writingseektable",q
                q.writeseek(self.file,self.noseek) # APPENDED
    def finalize(self):          
        if not self.endemitted:
            writeend(self.file)       # APPENDED TWICE
            self.endemitted = True
        #TODO remove device node if not emitted
        #removed {'rt': 7, 'ps': 0, 'fs': 28, 'poffset': 368218784, 'hoffset': 368218756, 'nid': 2, 'nextheader': 368218784, 'undopos': 1873}
        for q in self.stats.values():
            if not q.emitted:
                #print "writingseektable",q
                if not q.removeemitted:
                    #TODO remove this if not emitted
                    #REMOVE is EMPTY
                    #NodeRemovedRecord record(m_pRecordBuffer, RECORD_MAX_SIZE, FALSE);
                    #record.SetNodeID(recordedNodeInfo.nNodeID);
                    #record.SetUndoRecordPos(recordedNodeInfo.nNodeAddedPos);                    
                    pass
                q.writeseek(self.file,self.noseek) # APPENDED
        # patch
        for q in self.stats.values():
            q.fixnodeadded(self.file) # seeks to location

        # patch the head
        self.h0["maxnid"]   = self.mid
        self.h0["ts"] = max([q.maxts for q in self.stats.values()])
        if self.h0["ts"] is None:
            self.h0["ts"] = 0
            print "WARNING issue with stats",self.stats
        writehead1(self.file,self.h0)       

#                    writehead(b,hc)
#                    b.write(colordata)


#                nidc = r.nodetype2nid[NODE_TYPE_IMAGE]
#               nidd = r.nodetype2nid[NODE_TYPE_DEPTH]
#                q = stats[nidc]
#                q.addframe(h,hh,b)              
#                q = stats[nidd]
#                q.addframe(h,hh,b)



from . import onifile as oni
import struct,array,os,sys
try:
    from ..xndec import xndec as xndec 
except:
    xndec = None
try:
    import png
except:
    png =None

def anyframeextract(r,a,h,pd,codec,extractpath,args,useframeid=False):
    q = r.streams[h["nid"]]
    xres,yres = q["size"]
    if codec == "jpeg":
        ext = "jpeg"
    elif codec == "16zt":
        ext = "png"
        if ob is None:
            ob = xndec.allocoutput16(xres*yres)
    else:
        ext = "bin"
    if useframeid:
        outfile = "%s%d.%s" % (extractpath,pd["frameid"],ext)
    else:
        outfile = "%s.%s" % (extractpath,ext)
    print outfile,codec
    of = open(outfile,"wb")
    if codec == "jpeg":
        of.write(a.read(h["ps"]))
    elif codec == "16zt":
        code,size = xndec.doXnStreamUncompressDepth16ZWithEmbTable(a.read(h["ps"]),ob)
        aa = array.array("H")
        aa.fromstring(ob)
        rows = [aa[i*xres:(i+1)*xres] for i in range(0,yres)]
        if args.coloreddepth:                                    
            # save content of ob to PNG 16bit with size xres,yres
            w = png.Writer(width=xres,height=yres,greyscale=True,bitdepth=8)
            w.write(of,[[int(x*255.0/8000.0) for x in y] for y in rows])
        else:
            # save content of ob to PNG 16bit with size xres,yres
            w = png.Writer(width=xres,height=yres,greyscale=True,bitdepth=16)
            w.write(of,rows)
    else:
        print "unsupported codec",foundhh
        sys.exit(0)
    of.close()

def extractframes(args,a,frames,outpudir):
    """Extract the frames from the a file using the lisitng"""
    cd = os.path.join(outpudir,"color")
    ird  = os.path.join(outpudir,"ir")
    if not os.path.isdir(cd):
        os.makedirs(cd)
    if not os.path.isdir(ird):
        os.makedirs(ird)
    r = oni.Reader(a)
    ctargetnid = None
    irtargetnid = None
    ccodec = None
    ircodec = None
    nextir = False
    idx = 0
    nextidx = 0
    while True:
        h = r.next()
        if h is None:
            break
        if h["rt"] == oni.RECORD_NODE_ADDED:
            hh = oni.parseadded(a,h)
            if hh["nodetype"] == oni.NODE_TYPE_IMAGE:
                ctargetnid = h["nid"]
                ccodec = hh["codec"]
            elif hh["nodetype"] == oni.NODE_TYPE_IR:
                print "found IR node type"
                ircodec = h["codec"]
                irtargetnid = hh["nid"]
        elif h["nid"] == ctargetnid:
            if h["rt"] == oni.RECORD_NEW_DATA:
                pd = oni.parsedatahead(a,h)
                if idx >= len(frames):
                    break
                if pd["frameid"] == frames[idx]:
                    anyframeextract(r,a,h,pd,ccodec,os.path.join(cd,"frame_%d" % idx),args)
                    # extract id as idx
                    if irtargetnid is not None:
                        nextir = True
                        nextidx = idx
                    idx = idx + 1
        elif h["nid"] == irtargetnid and nextir:
            if h["rt"] == oni.RECORD_NEW_DATA:
                pd = oni.parsedatahead(a,h)
                anyframeextract(r,a,h,pd,ircodec,os.path.join(cd,"frame_%d" % nextidx),args)
                nextir = False
                if idx >= len(frames):
                    break
def extractmjpeg(args,a,b):
    targetnid = None
    seekframe = args.fseek
    endframe = args.fduration < 0 and -1 or seekframe+args.fduration
    # scan all and keep pre and last
    r = oni.Reader(a)
    while True:
        h = r.next()
        if h is None:
            break
        if h["rt"] == oni.RECORD_NODE_ADDED:
            hh = oni.parseadded(a,h)
            if hh["nodetype"] == oni.NODE_TYPE_IMAGE:
                targetnid = h["nid"]
        elif h["nid"] == targetnid:
            if h["rt"] == oni.RECORD_NEW_DATA:
                pd = oni.parsedatahead(a,h)
                if pd["frameid"] >= seekframe and (endframe == -1 or pd["frameid"] < endframe):
                    print pd["frameid"],h["ps"],h["fs"]
                    b.write(a.read(h["ps"]))         
                elif endframe != -1 and pd["frameid"] >= endframe:
                    break

# TODO extract both
def extract(args,action,extractpath,a):
    codec = None
    foundhh = None
    ob = None
    extracttype = dict(extractboth=-1,extractcolor=oni.NODE_TYPE_IMAGE,extractdepth=oni.NODE_TYPE_DEPTH,extractir=oni.NODE_TYPE_IR)[action]
    targetnid = None

    seekframe = args.fseek
    endframe = args.fduration < 0 and -1 or seekframe+args.fduration
    # scan all and keep pre and last
    r = oni.Reader(a)
    while True:
        h = r.next()
        if h is None:
            break
        if h["rt"] == oni.RECORD_NODE_ADDED:
            hh = oni.parseadded(a,h)
            if hh["nodetype"] == extracttype:
                targetnid = h["nid"]
                codec = hh["codec"]
                foundhh = hh
                print "!!found matching for ",targetnid,codec
                if codec == "16zt":
                    if xndec is None or xndec.doXnStreamUncompressDepth16ZWithEmbTable is None:
                        print "xndec is missing cannot decode to png"
                        sys.exit(-1)
                    if png is None:
                        print "pypng is missing"
                        sys.exit(-1)
        elif h["nid"] == targetnid:
            if h["rt"] == oni.RECORD_NEW_DATA:
                pd = oni.parsedatahead(a,h)
                if pd["frameid"] >= seekframe and (endframe == -1 or pd["frameid"] < endframe):
                    print "newdata",pd["frameid"],h["ps"],h["fs"],codec
                    q = r.streams[h["nid"]]
                    xres,yres = q["size"]
                    if codec == "jpeg":
                        ext = "jpeg"
                    elif codec == "16zt":
                        ext = "png"
                        if ob is None:
                            ob = xndec.allocoutput16(xres*yres)
                    else:
                        ext = "bin"
                    outfile = "%s%d.%s" % (extractpath,pd["frameid"],ext)
                    print outfile,codec
                    of = open(outfile,"wb")
                    if codec == "jpeg":
                        of.write(a.read(h["ps"]))
                    elif codec == "16zt":
                        code,size = xndec.doXnStreamUncompressDepth16ZWithEmbTable(a.read(h["ps"]),ob)
                        aa = array.array("H")
                        aa.fromstring(ob)
                        rows = [aa[i*xres:(i+1)*xres] for i in range(0,yres)]
                        if args.coloreddepth:                                    
                            # save content of ob to PNG 16bit with size xres,yres
                            w = png.Writer(width=xres,height=yres,greyscale=True,bitdepth=8)
                            w.write(of,[[int(x*255.0/8000.0) for x in y] for y in rows])
                        else:
                            # save content of ob to PNG 16bit with size xres,yres
                            w = png.Writer(width=xres,height=yres,greyscale=True,bitdepth=16)
                            w.write(of,rows)
                    else:
                        print "unsupported codec",foundhh
                        sys.exit(0)
                    of.close()
                elif endframe != -1 and pd["frameid"] >= endframe:
                    break


from . import onifile as oni
import kinect1 as device
import struct
from collections import defaultdict

def fixnite(args,action,a,b):
    r = oni.Reader(a)
    w = oni.Writer(b,r.h0)
    if args.noseek:
        w.noseek = True
    ctargetnid = None
    ctargetnid2 = None
    emitted = False
    emitted2 = False
    foundprop = set()
    foundprop2 = set()
    neededprop = device.niteprops()
    neededprop.update(dict(IsFrameBased=1,IsPixelBased=1,IsStream=1,FrameSync=1,xnIsGenerating=1))
    neededprop2 = dict(IsFrameBased=1,IsPixelBased=1,IsStream=1,FrameSync=0,xnIsGenerating=1)
    if args.registered != -1:
        neededprop["RegistrationType"] = args.registered
        #XN_STREAM_PROPERTY_REGISTRATION_TYPE = 0x10801005, // "RegistrationType"
    print "fixing props",neededprop
    while True:
        h = r.next()
        if h is None:
            break
        if ctargetnid is not None and h["nid"] != ctargetnid and not emitted:
            # inject ParamCoeff = 4 for depth
            for k,v in neededprop.iteritems():
                if k in foundprop:
                    print "already present",k
                else:
                    print "adding property",k
                    if k == "ZPPS" or k == "LDDIS":
                        w.addprop(ctargetnid,k,oni.RECORD_REAL_PROPERTY,v)                        
                    elif k == "S2D":
                        n = device.S2D().tostring()
                        if len(n) != 4096: # 
                            print "ERROR! in S2D"
                            sys.exit(0)
                        w.addprop(ctargetnid,k,oni.RECORD_GENERAL_PROPERTY,n)
                    elif k == "D2S":
                        n = device.D2S().tostring()
                        if len(n) != 20002:
                            print "ERROR! in S2D"
                            sys.exit(0)
                        w.addprop(ctargetnid,k,oni.RECORD_GENERAL_PROPERTY,n)
                    else:
                        w.addprop(ctargetnid,k,oni.RECORD_INT_PROPERTY,v)
            #XN_STREAM_PROPERTY_S2D_TABLE S2D GENERAL 4096
            #XN_STREAM_PROPERTY_D2S_TABLE D2S GENERAL 20002
            emitted = True
        if ctargetnid2 is not None and h["nid"] != ctargetnid2 and not emitted2:
            # inject ParamCoeff = 4 for depth
            for k,v in neededprop2.iteritems():
                if k in foundprop2:
                    print "already present",k
                else:
                    print "adding property",k
                    w.addprop(ctargetnid2,k,oni.RECORD_INT_PROPERTY,v)
            #XN_STREAM_PROPERTY_S2D_TABLE S2D GENERAL 4096
            #XN_STREAM_PROPERTY_D2S_TABLE D2S GENERAL 20002
            emitted2 = True        
        if ctargetnid == h["nid"] and h["rt"] == oni.RECORD_INT_PROPERTY:
            p = oni.parseprop(a,h)
            if args.registered != -1 and p["name"] == "RegistrationType":
                print "replacing existing RegistrationType"
                w.addprop(ctargetnid,p["name"],oni.RECORD_INT_PROPERTY,args.registered)                            
            else:
                w.copyblock(h,a)
            foundprop.add(p["name"])
        elif ctargetnid == h["nid"] and h["rt"] == oni.RECORD_GENERAL_PROPERTY:
            p = oni.parseprop(a,h)
            foundprop.add(p["name"])
            w.copyblock(h,a)
        elif ctargetnid2 == h["nid"] and h["rt"] == oni.RECORD_INT_PROPERTY:
            p = oni.parseprop(a,h)
            foundprop2.add(p["name"])
            w.copyblock(h,a)
        elif h["rt"] == oni.RECORD_NODE_ADDED:
            hh = oni.parseadded(a,h)
            if hh["nodetype"] == oni.NODE_TYPE_DEPTH:
                ctargetnid = h["nid"]
                print "found depth",ctargetnid
            elif hh["nodetype"] == oni.NODE_TYPE_IMAGE:
                ctargetnid2 = h["nid"]
                print "found color",ctargetnid2
            w.copyblock(h,a)
        elif h["rt"] == oni.RECORD_SEEK_TABLE:
            w.emitseek(h["nid"]) # we are changing props
        else:
            w.copyblock(h,a)
    w.finalize()


def makeregistered(args,action,a,b):
    r = oni.Reader(a)
    w = oni.Writer(b,r.h0)
    if args.noseek:
        w.noseek = True
    ctargetnid = None
    emitted = False
    foundprop = set()
    neededprop = {}
    already = False
    neededprop["RegistrationType"] = 2 # default
    if args.registered != -1:
        neededprop["RegistrationType"] = args.registered
    while True:
        h = r.next()
        if h is None:
            break
        if ctargetnid is not None and h["nid"] != ctargetnid and not emitted:
            # inject ParamCoeff = 4 for depth
            for k,v in neededprop.iteritems():
                if k in foundprop:
                    print "already present",k
                else:
                    print "adding property",k,"as",neededprop["RegistrationType"]
                    w.addprop(ctargetnid,k,oni.RECORD_INT_PROPERTY,v)
            emitted = True
        if ctargetnid == h["nid"] and h["rt"] == oni.RECORD_INT_PROPERTY:
            p = oni.parseprop(a,h)
            if p["name"] == "RegistrationType":
                print "replacing existing RegistrationType ",p["data"], " with ",neededprop["RegistrationType"]
                w.addprop(ctargetnid,p["name"],oni.RECORD_INT_PROPERTY,neededprop["RegistrationType"],datalen=p["datalen"])
                already = True
            else:
                w.copyblock(h,a)
            foundprop.add(p["name"])
        elif h["rt"] == oni.RECORD_NODE_ADDED:
            hh = oni.parseadded(a,h)
            if hh["nodetype"] == oni.NODE_TYPE_DEPTH:
                ctargetnid = h["nid"]
                print "found depth",ctargetnid
            w.copyblock(h,a)
        elif h["rt"] == oni.RECORD_SEEK_TABLE:
            if already: # needed for compare
                w.emitseek(h["nid"],a,h)
            else:
                w.emitseek(h["nid"])
        else:
            w.copyblock(h,a)
    w.finalize()

def fix(args,action,a):
    r = oni.Patcher(a)
    last = None
    prelast = None	
    mid = -1
    needclose = False
    print "fileheader",r.h0
    while True:
        h = r.next()
        if h is None:
            break
        elif h.magic != oni.magic:
            break
        elif h["rt"] == oni.RECORD_NEW_DATA:
            break
        if mid < h["nid"]:
            mid = h["nid"]
        prelast = last
        last = h
        if h["rt"] == oni.RECORD_NEW_DATA:
            hh = oni.parsedatahead(a,h)
            q = r.stats[h["nid"]]
            q.addframe(h,hh)
           
    if True:
        print "prelast",prelast
        print "last",last
        print "maxnodeid",mid

    if not last["nextheader"] <= filesize:
        print "FIXER last ends at",last["nextheader"],"with file size",filesize,"delta:",last["nextheader"]-filesize," we need truncate at pre last",prelast["afteroffset"]
        if action == "fixcut":
            a.truncate(prelast["nextheader"])
        last = prelast
        needclose = True

    if last["rt"] != oni.RECORD_END:
        print "FIXER missing RECORD_END. Last Record is:",last["rt"],"appending"
        if action == "fixcut":
            a.seek(0,2)
            oni.writeend(a)
            needclose = True

    if h0["maxnid"] != mid:
        print "FIXER bad maxnid",h0["maxnid"],"vs",mid,"fixing"
        h0["maxnid"] = mid
        if action == "fixcut":
            a.seek(0,0)
            oni.writehead1(a,h0)
            needclose = True
    else:
        print "header ok, no changes"
    if needclose:
    	a.finalize()



from . import onifile as oni
import struct
from collections import defaultdict

def copy(args,a,b):
    r = oni.Reader(a)
    w = oni.Writer(b,r.h0)
    while True:
        h = r.next()
        if h is None:
            break
        elif h["rt"] == oni.RECORD_SEEK_TABLE:
            w.emitseek(h["nid"])
        elif h["rt"] == oni.RECORD_NEW_DATA:
            hh = oni.parsedatahead(a,h)            
            print dict(nid=h["nid"],ps=h["ps"],fs=h["fs"],frameid=hh["frameid"],timestamp=hh["timestamp"])
            w.addframe(h["nid"],hh["frameid"],hh["timestamp"],a.read(h["ps"]))
        else:
            w.copyblock(h,a)

    w.finalize()

def cut(args,target,a,b):
    r = oni.Reader(a)
    w = oni.Writer(b,r.h0)
    qf = dict()
    while True:
        h = r.next()
        if h is None:
            break
        elif h["rt"] == oni.RECORD_SEEK_TABLE:
            w.emitseek(h["nid"])
        elif h["rt"] == oni.RECORD_NEW_DATA:
            hh = oni.parsedatahead(a,h)            
            # check timestamp/frame
            if target[0] == "frame":
                good = hh["frameid"] >= target[1][0] and hh["frameid"] <= target[1][1]
            else:
                good = hh["timestamp"] >= target[1][0] and hh["timestamp"] <= target[1][1]
            if good:
                qff = qf.get(h["nid"])
                if qff is None:
                    qff = hh # original
                    qf[h["nid"]] = qff
                print dict(nid=h["nid"],ps=h["ps"],fs=h["fs"],frameid=hh["frameid"],timestamp=hh["timestamp"])
                w.addframe(h["nid"],hh["frameid"]-qff["frameid"]+1,hh["timestamp"]-qff["timestamp"],a.read(h["ps"]))
        else:
            w.copyblock(h,a)
    w.finalize()


def strip(args,action,a,b):
    # scan all and keep pre and last
    r = oni.Reader(a)
    w = oni.Writer(b)

    ignoredtype = dict(stripcolor=oni.NODE_TYPE_IMAGE,stripdepth=oni.NODE_TYPE_DEPTH,stripir=oni.NODE_TYPE_IR)[action]
    ignoredid = -1
    while True:
        h = r.next()
        if h is None:
            break
        elif h["rt"] == oni.RECORD_NODE_ADDED:
            hh = oni.parseadded(a,h)
            if hh["nodetype"] == ignoredtype:
                ignoredid = h["nid"]
            else:
                w.copyblock(h,a)
        elif h["nid"] == ignoredid:
            continue
        elif h["rt"] == oni.RECORD_SEEK_TABLE:
            w.emitseek(h["nid"])
        elif h["rt"] == oni.RECORD_NEW_DATA:
            hh = oni.parsedatahead(a,h)            
            print dict(nid=h["nid"],ps=h["ps"],fs=h["fs"],frameid=hh["frameid"],timestamp=hh["timestamp"])
            w.addframe(h["nid"],hh["frameid"],hh["timestamp"],a.read(h["ps"]))
        else:
            w.copyblock(h,a)
    w.finalize()


def skip(args,a,b):
    r = oni.Reader(a)
    w = oni.Writer(b)
    qf = defaultdict(lambda: 0)
    while True:
        h = r.next()
        if h is None:
            break
        elif h["rt"] == oni.RECORD_SEEK_TABLE:
            w.emitseek(h["nid"])
        elif h["rt"] == oni.RECORD_NEW_DATA:
            hh = oni.parsedatahead(a,h)
            oldframe = qf[h["nid"]]
            if (oldframe % args.skipframes) == 0:
                w.addframe(h["nid"],hh["frameid"],hh["timestamp"],a.read(h["ps"]))
            qf[h["nid"]] += 1
        else:
            w.copyblock(h,a)
    w.finalize()

def dupframes(args,a,b):
    r = oni.Reader(a)
    w = oni.Writer(b)
    dt = 33333 #ms virtual
    qf = dict()
    while True:
        h = r.next()
        if h is None:
            break
        elif h["rt"] == oni.RECORD_SEEK_TABLE:
            w.emitseek(h["nid"])
        elif h["rt"] == oni.RECORD_NEW_DATA:
            hh = oni.parsedatahead(a,h)
            dd = a.read(h["ps"])
            qff = qf.get(h["nid"])
            if qff is None:
                qff = dict(frameid=hh["frameid"],timestamp=hh["timestamp"])
                qf[h["nid"]] = qff
            for i in range(0,args.dupframes):
                w.addframe(h["nid"],qff["frameid"],qff["timestamp"],dd)
                qff["frameid"] += 1
                qff["timestamp"] += dt
        else:
            w.copyblock(h,a)
    w.finalize()





import ctypes,sys,png,array,os
from PIL import Image

#	void register2color(char * outrgb, int cw, int ch, const char * rgb, int dw, int dh, const uint16_t * depth, const double depthK[9], const double colorK[9], const double inrotation[9], const double position[3])
#	void register2depth(uint16_t * outdepth, int cw, int ch, const char * rgb, int dw, int dh, const uint16_t * depth, const double depthK[9], const double colorK[9], const double inrotation[9], const double position[3])

p8 = ctypes.POINTER(ctypes.c_uint8) # for rgb
p16 = ctypes.POINTER(ctypes.c_uint16) # for depth
pd = ctypes.POINTER(ctypes.c_double) # generic
ad3 = ctypes.c_double * 3  # for pos
ad5 = ctypes.c_double * 5  # for dist
ad9 = ctypes.c_double * 9  # for intrinsics
ad16 = ctypes.c_double * 16 # for affine

ci = ctypes.c_int;

types = ["",".so",".dylib",".pyd"]
x = None
path = os.path.abspath(__file__)
dir_path = os.path.dirname(path)

for a in types:
	try:
		x = ctypes.CDLL(os.path.join(dir_path,"libanyregistration" + a))
		break
	except:
		pass

f = x.register2color
f.argtypes = [p8,ci,ci,p8,ci,ci,p16,ad9,ad9,ad9,ad3]
register2color = f

f = x.register2depth
f.argtypes = [p16,ci,ci,p8,ci,ci,p16,ad9,ad9,ad9,ad3]
register2depth = f

def depth2color(x):
	c = min(255,int(x*255.0/2000.0))
	if c != 0:
		c = 255-c;
	return (c,c,0)

def blend(p,i,c):
	a = 0.7
	b = 0.3
	p[i*3] = int(p[i*3]*a + c[0]*b)
	p[i*3+1] = int(p[i*3+1]*a  + c[1]*b)
	p[i*3+2] = int(p[i*3+2]*a + c[2]*b)


def image2buffer(name):
	# http://pillow.readthedocs.org/en/latest/reference/Image.html
	inc_i = Image.open(name)
	inc = inc_i.tostring() #Image.tobytes(inc_i,encoder_name='raw')
	return inc

def buffer2image(buffer,size):
	return Image.frombytes("RGB", size, buffer, decoder_name='raw')

def saveimage(image,name,t="JPEG"):
	print "saving image",image
	image.save(name,t)

def savedepthpng(buffer,size,name):
	xres,yres = size
	w = png.Writer(width=size[0],height=size[1],greyscale=True,bitdepth=16)
	aa = array.array("H")
	aa.fromstring(buffer)
	print "saving",size,"of",len(aa)
	w.write(open(name,"wb"),[aa[i*xres:(i+1)*xres] for i in range(0,yres)])	

def savedepthpngcolor(buffer,size,name):
	xres,yres = size
	w = png.Writer(width=size[0],height=size[1],greyscale=True,bitdepth=8)
	aa = array.array("H")
	aa.fromstring(buffer)
	print "saving",size,"of",len(aa)
	rows = [aa[i*xres:(i+1)*xres] for i in range(0,yres)]
	w.write(open(name,"wb"),[[min(int(x*255.0/8000.0),255) for x in y] for y in rows])

def depthpng2buffer(name):
	# load png filename grayscale 16 bit
	p = png.Reader(open(name,"rb"))
	w,h,pi,me = p.read_flat()
	print "read ",w,h,len(pi),me
	return pi

def savedepth(buffer,name):
	open(name,"wb").write(buffer)

def mergedepthcolor(d,c):
	ad = array.array("H")
	ad.fromstring(d)
	aa = array.array("B")
	aa.fromstring(c)
	for i in range(0,len(ad)):
		x = ad[i]
		blend(aa,i,depth2color(ad[i]))
	return aa.tostring()

def doregister2color(outc,depthin,colorin,sizergb,Drgb,Krgb,sizedepth,Ddepth,Kdepth,R,T):
	if outc is None:
		outc = ctypes.create_string_buffer(sizedepth[0]*sizedepth[1]*3)
	poutc = ctypes.cast(outc,p8)
	prgb = ctypes.cast(ctypes.c_char_p(colorin),p8)
	if isinstance(depthin,array.array):
		print "input is array"
		addr, count = depthin.buffer_info()
		pdepth = ctypes.cast(addr, p16)
	else:
		pdepth = ctypes.cast(depthin,p16)
	pKdepth = ad9(*Kdepth)
	pDdepth = ad5(*Ddepth)
	pKrgb = ad9(*Krgb)
	pDrgb = ad5(*Drgb)
	pR = ad9(*R)
	pT = ad3(*T)
	register2color(poutc,sizergb[0],sizergb[1],prgb,sizedepth[0],sizedepth[1],pdepth,pKdepth,pKrgb,pR,pT)
	return outc

def doregister2depth(outd,depthin,colorin,sizergb,Drgb,Krgb,sizedepth,Ddepth,Kdepth,R,T):
	if outd is None:
		outd = ctypes.create_string_buffer(sizergb[0]*sizergb[1]*2)
	#outd = (ctypes.c_ushort * (sizedepth[0]*sizedepth[1]))()
	poutd = ctypes.cast(outd,p16)
	prgb = ctypes.cast(ctypes.c_char_p(colorin),p8)
	if isinstance(depthin,array.array):
		addr, count = depthin.buffer_info()
		pdepth = ctypes.cast(addr, p16)
	else:
		pdepth = ctypes.cast(ctypes.c_char_p(depthin),p16)
	pKdepth = ad9(*Kdepth)
	pDdepth = ad5(*Ddepth)
	pKrgb = ad9(*Krgb)
	pDrgb = ad5(*Drgb)
	pR = ad9(*R)
	pT = ad3(*T)
	print ctypes.addressof(poutd.contents)
	register2depth(poutd,sizergb[0],sizergb[1],prgb,sizedepth[0],sizedepth[1],pdepth,pKdepth,pKrgb,pR,pT)
	return outd

#TODO merge the merged images and output ad jpeg or png to show alignment

if __name__ == '__main__':

	# Asus xtion, distortion is unused
	size = (640,480)
	Drgb = (0.041623, -0.105707, -0.001538, 0.000755, 0.0)
	Ddepth = (0.041623, -0.105707, -0.001538, 0.000755, 0.0)

	Krgb = (533.9089, 0.0, 321.09951,    0.0, 534.555793, 237.786129,    0.0, 0.0, 1.0)
	Kdepth = (570, 0.0, 320,   0.0, 570, 240,    0.0, 0.0, 1.0)


	#http://qianyi.info/scenedata.html

	#http://robotics.dei.unipd.it/images/Papers/Conferences/basso2014unsupervised.pdf
	q= (0.999983,-0.00337542, 0.00425487,0.00190883)
	R = (1,0,0,  0,1,0, 0,0,1)
	T = (0.0222252,0,0)
	#export A=/Users/eruffaldi/Desktop/scotachtura2.oni
	#./pyonitool.py --extractdepth out --fseek 0 --fduration 1 $A
	#./pyonitool.py --extractcolor out --fseek 0 --fduration 1 $A

	if len(sys.argv) == 3:
		indepth = sys.argv[1]
		incolor = sys.argv[2]
	else:
		indepth = "depth.png"
		incolor = "color.jpg"
	if indepth.endswith(".png"):
		ind = depthpng2buffer(indepth) # note that pillow does not sypport 16bit
	else:
		ind = open(indepth,"rb").read()
	inc = image2buffer(incolor)

	outc = doregister2color(None,ind,inc,size,Drgb,Krgb,size,Ddepth,Kdepth,R,T)
	outd = doregister2depth(None,ind,inc,size,Drgb,Krgb,size,Ddepth,Kdepth,R,T)

	if indepth.endswith(".png"):
		savedepthpng(outd.raw,size,"outdepth16.png")
		savedepthpngcolor(outd.raw,size,"outdepthc.png")
	else:
		savedepth(outd.raw,"outdepth.bin")

	outci = buffer2image(outc.raw,size)
	saveimage(outci,"outimage.jpg")

	outci = buffer2image(mergedepthcolor(ind,inc),size)
	saveimage(outci,"mergeinput.jpg")

	outci = buffer2image(mergedepthcolor(ind,outc),size)
	saveimage(outci,"mergenewcolor.jpg")

	outci = buffer2image(mergedepthcolor(outd,inc),size)
	saveimage(outci,"mergenewdepth.jpg")

	# merge the images: inputdepth+adjustedcolor OR inputcolor+adjusteddepth



import struct#
from . import onifile as oni
import shutil,io
import array,os
import bisect
from collections import defaultdict
from anyregistration import anyregistration
from PIL import Image
from xndec import xndec as xndec
try:
    import png
except:
    png =None


class AsusModel:
    def __init__(self):
        self.size = (640,480)
        self.Drgb = (0.041623, -0.105707, -0.001538, 0.000755, 0.0)
        self.Ddepth = (0.041623, -0.105707, -0.001538, 0.000755, 0.0)
        self.Krgb = (533.9089, 0.0, 321.09951,    0.0, 534.555793, 237.786129,    0.0, 0.0, 1.0)
        self.Kdepth = (570, 0.0, 320,   0.0, 570, 240,    0.0, 0.0, 1.0)
        self.R = (1,0,0,  0,1,0, 0,0,1)
        self.T = (0.0222252,0,0)

def register(args,action,a,b):

    model = AsusModel()
	# scan first and emit second
    depthseen = False
    colorseen = False
    device = "asusxtion"
    ob = None
    depthdataout = None
    colordataout = None
    maxoutdepth = 5000

    r = oni.Reader(a)
    w = oni.Writer(b,r.h0)

    modifycolor = action == "registercolor" # color from depth or the opposite, means that we can directly store one of the two
    nidd = None
    nidc = None
    # parse all metadata and copy
    ignore = set([oni.RECORD_NEW_DATA,oni.RECORD_END,oni.RECORD_SEEK_TABLE,oni.RECORD_NODE_REMOVED])
    while True:
        h = r.next()
        if h is None:
            break
        elif h["rt"] == oni.RECORD_NEW_DATA:
            continue # otherwise no seek
        elif h["rt"] in ignore:
            print "ignored",h["rt"]
            continue
        #TODO mark registered
        #        elif h["rt"] == oni.RECORD_INT_PROPERTY:
        #            z = oni.parseprop(a,h)
        #            print z["name"]
        #            if z["name"] == "registrationType":
        #                z["data"] = 2
        #            # TODO append block property
        #            #w.addproperty(h,z)        else:
        else:
            w.copyblock(h,a)
    # then start iterating the frames
    idd = r.nodetype2nid[oni.NODE_TYPE_DEPTH]
    idc = r.nodetype2nid[oni.NODE_TYPE_IMAGE]
    td = r.getseektable(idd)
    tc = r.getseektable(idc)
    if tc is None:
        raise "Missing color"
    if td is None:
        raise "Missing depth"        
    tc["nid"] = idc
    td["nid"] = idd
    tc["times"] = [x["timestamp"] for x in tc["data"]]
    td["times"] = [x["timestamp"] for x in td["data"]]

    def findnearest(index,time):
        a = index["times"]
        i = bisect.bisect_left(a,time)
        # time <= a[i]
        if i == 0:
            if len(a) < 2:
                return 0
            else:
                i = 1 # 0 is bad
        if i < 2 or i >= len(a)-1:
            if i >= len(a)-1:
                i = len(a)-1
            return (i,index["data"][i])
        else:
            next = a[i] 
            pre = a[i-1]
            if (next-time) < (time-pre):
                return (i,index["data"][i])
            else:
                return (i-1,index["data"][i-1])


    # given the primary 
    if modifycolor:
        primary = td
        secondary = tc
    else:
        primary = tc
        secondary = td

    # statistics
    if False:
        dtimes = []
        dframes = []
        for x in primary["data"]:
            fi,f = findnearest(secondary,x["timestamp"])
            dtimes.append(f["timestamp"]-x["timestamp"])
            dframes.append(f["frameid"]-x["frameid"])
            print x["frameid"],f["frameid"],x["timestamp"],f["timestamp"],(f["frameid"]-x["frameid"]),(f["timestamp"]-x["timestamp"]),(f["offset"]-x["offset"])
        print "time min",min(dtimes),"max",max(dtimes)
        print "frame min",min(dframes),"max",max(dframes)

    for primaryframe in primary["data"]:
        fi,secondaryframe = findnearest(secondary,primaryframe["timestamp"])

        # TODO: support for by frame matching, simply: f = secondary[x["frameid"]]
        if args.fduration > -1 and primaryframe["frameid"] > args.fduration:
            break

        # skip dummy first
        if primaryframe["frameid"] == 0 or secondaryframe["frameid"] == 0:
            continue

        for i in range(0,2): #maybe just (primaryframe,secondaryframe)
            hh = i == 0 and primaryframe or secondaryframe

            a.seek(hh["offset"]) # to data
            h = oni.readrechead(a)
            hhreal = oni.parsedatahead(a,h)
            print h,hhreal
            nid = h["nid"]

            # TODO support for sizes different among streams
            hhnode = r.streams[nid]
            codec = hhnode["codec"]            
            size = hhnode["size"]
            xres,yres = size

            #TODO allow for other encodings
            if codec == "jpeg":
                rawdata = a.read(h["ps"])             
                # decompress always
                b = io.BytesIO()
                b.write(rawdata)
                b.seek(0)
                inc_i = Image.open(b)

                # AS COLOR
                hc = hh
                if not modifycolor:
                    # preserve color as compressed
                    w.addframe(nid,hh["frameid"],hh["timestamp"],rawdata)
                colordata = inc_i.tostring() #Image.tobytes(inc_i,encoder_name='raw')
            elif codec == "16zt":
                rawdata = a.read(h["ps"])
                if ob is None:
                    ob = xndec.allocoutput16(xres*yres)
                xndec.doXnStreamUncompressDepth16ZWithEmbTable(rawdata,ob)
                aa = array.array("H")
                aa.fromstring(ob)

                # AS DEPTH
                depthdata = aa
                hd = hh
                if modifycolor:
                    # depth is preserved in depthmode
                    w.addframe(nid,hh["frameid"],hh["timestamp"],rawdata)
            else:
                print "unsupported codec",codec


        print "registering ",modifycolor and "newcolor" or "newdepth","depthframe(frame,time)",(hd["frameid"],hd["timestamp"]),"colorframe(frame,time)",(hc["frameid"],hc["timestamp"]),"deltaframe(c-d)(frame,time)",(hc["frameid"]-hd["frameid"],hc["timestamp"]-hd["timestamp"])
        if modifycolor:
            #anyregistration on (colordata,depthdata) -> colordataout
            # autoalloc
            print depthdata.__class__,len(depthdata),colordata.__class__,len(colordata),size
            colordataout = anyregistration.doregister2color(colordataout,depthdata,colordata,size,model.Drgb,model.Krgb,size,model.Ddepth,model.Kdepth,model.R,model.T)
            im = Image.frombytes("RGB", size, colordataout.raw, decoder_name='raw')
            b = io.BytesIO()
            im.save(b, 'JPEG')
            w.addframe(idc,hd["frameid"],args.registersynctime and hd["timestamp"] or hc["timestamp"],b.getvalue()) # keep timestamp, discard frameid
            print "\tcompressed from",xres*yres*3,"to",len(b.getvalue())
            # TODO add option for faking timestamp
        else:
            #anyregistration (colordata,depthdata) -> depthdataout
            depthdataout = anyregistration.doregister2depth(depthdataout,depthdata,colordata,size,model.Drgb,model.Krgb,size,model.Ddepth,model.Kdepth,model.R,model.T)
            print "\tgenerated",depthdataout.__class__,len(depthdataout)
            status,size = xndec.doXnStreamCompressDepth16ZWithEmbTable(depthdataout,ob,maxoutdepth)
            print "\tcompressed from",xres*yres*2,"to",size
            w.addframe(idd,hc["frameid"],args.registersynctime and hc["timestamp"] or hd["timestamp"],ob[0:size]) # keep timestamp, discard frameid
            # TODO add option for faking timestamp

    #TODO
    #end
    #removed DEVICE  --> MISSING
    #removed IMAGE   --> 
    #seektable IMAGE
    #removed DEPTH   --> 
    #seektable DEPTH
    w.finalize()


